#include <Arduino.h>

// Obsluha weboveho serveru
Ticker deferred; // zpozdene veci (konkretne restart ESP treba po aktualizaci)
const char TEXTPLAIN[] PROGMEM = "text/plain";
const char TEXTJSON[] PROGMEM = "text/json";
const char TEXTHTML[] PROGMEM = "text/html";
const char CCLEAR[] PROGMEM = "Crash area cleared!";

const char RESTARTING[] PROGMEM = "Restarting...";

void handleReset(AsyncWebServerRequest *request)
{

  deferred.once_ms(500, []()
  {
    ESP.restart();
  });

  request->send_P(200, FPSTR(TEXTPLAIN), RESTARTING);
}

void handleEEDump(AsyncWebServerRequest *request)
{
  AsyncWebServerResponse *response = request->beginChunkedResponse("application/octet-stream", [](uint8_t *buffer, size_t maxLen, size_t index) -> size_t {
    size_t written = 0;

    while ((index < EEPROM_SIZE) && (written < maxLen))
    {
      buffer[written] = EEPROM.read(index);
      ++index;
      ++written;
    }
    return written;
  });
  request->send(response);
}

#ifdef USE_CRASHDUMP
void handleCrash(AsyncWebServerRequest *request)
{
  String st;

  st.reserve(2000);
  crashGet(st);
  request->send(200, FPSTR(TEXTHTML), st);
}

void handleCrashClear(AsyncWebServerRequest *request)
{
  String st;

  crashClear();
  request->send_P(200, FPSTR(TEXTPLAIN), CCLEAR);
}
#endif

void handleNotFound(AsyncWebServerRequest *request)
{

#ifdef CORS_DEBUG
  if (request->method() == HTTP_OPTIONS)
  {
    request->send(200);
  }
  else
#endif
  {
    String message;

    message.reserve(1024);

    message = F("Page Not Found\r\n\r\nURI:");
    message.concat(request->url());
    message.concat(F("\r\nMethod: "));
    message.concat((request->method() == HTTP_GET) ? F("GET") : F("POST"));
    message.concat(F("\r\nParams: "));
    message.concat(request->params());
    message.concat(F("\r\n"));

    for (unsigned int i = 0; i < request->params(); i++ )
    {
      AsyncWebParameter *p = request->getParam(i);
      message.concat(F(" "));
      message.concat(p->name());
      message.concat(F(": "));
      message.concat(p->value());
      message.concat(F("\r\n"));
    }
    request->send(404, FPSTR(TEXTPLAIN), message);
  }
}

void www_onUpgrade(AsyncWebServerRequest *request)
{
  AsyncWebServerResponse *response = request->beginResponse(200, FPSTR(TEXTPLAIN), Update.hasError() ? F("FAIL") : F("OK"));

  response->addHeader(FPSTR("Connection"), FPSTR("close"));
  if (!Update.hasError())
  {
    deferred.once_ms(100, []()
      {
        ESP.restart();
      });
  }
  request->send(response);
}

void www_onUpgradeData(AsyncWebServerRequest *request, String filename, size_t index, uint8_t *data, size_t len, bool final)
{

    if (!index)
    {
        TRACE(TRACE_DEBUG, F("[UPGRADE] Start: %s"), filename.c_str());
        Update.runAsync(true);
        if (!Update.begin((ESP.getFreeSketchSpace() - 0x1000) & 0xFFFFF000))
        {
          TRACE(TRACE_ERROR, F("[UPGRADE] Error: %i"), Update.getError());
        }
    }
    if (!Update.hasError())
    {
        if (Update.write(data, len) != len)
        {
          TRACE(TRACE_ERROR, F("[UPGRADE] Error: %i"), Update.getError());
        }
    }
    if (final)
    {
        if (Update.end(true))
        {
            //TRACE(TRACE_DEBUG, F("[UPGRADE] Success:  %u bytes"), index + len);
        }
        else
        {
          TRACE(TRACE_ERROR, F("[UPGRADE] Error: %i"), Update.getError());
        }
    }
    else
    {
//        DEBUG_MSG_P(PSTR("[UPGRADE] Progress: %u bytes\r"), index + len);
    }
}

void www_onScan(AsyncWebServerRequest *request)
{
  String json;

  json.reserve(1024);
  json = F("[");
  int n = WiFi.scanComplete();

  if (n == -2)
  {
    WiFi.scanNetworks(true);
  } else if (n)
  {
    for (int i = 0; i < n; ++i)
    {
      if (i)
        json.concat(F(","));
      json.concat(("{\"rssi\":"));
      json.concat(String(WiFi.RSSI(i)));
      json.concat(F(",\"ssid\":\""));
      json.concat(WiFi.SSID(i));
      json.concat(F("\",\"bssid\":\""));
      json.concat(WiFi.BSSIDstr(i));
      json.concat(F("\",\"channel\":"));
      json.concat(String(WiFi.channel(i)));
      json.concat(F(",\"secure\":"));
      json.concat(String(WiFi.encryptionType(i)));
      json.concat(F(",\"hidden\":"));
      json.concat(String(WiFi.isHidden(i)?F("true"):F("false")));
      json.concat(F("}"));
    }
    WiFi.scanDelete();
    if (WiFi.scanComplete() == -2)
    {
      WiFi.scanNetworks(true);
    }
  }
  json.concat(F("]"));
  request->send(200, FPSTR(TEXTJSON), json);
}

void webserver_init(void)
{

  www.rewrite("/", "/index.htm");
  init_websocket();
#ifdef HTTP_UPDATE
  httpUpdater.setup(&www);
#endif
#ifdef USE_CRASHDUMP
  www.on("/crash", handleCrash);
  www.on("/crashclear", handleCrashClear);
#endif
  www.on("/rst", handleReset);
  www.on("/eedump", handleEEDump);
#ifdef USE_EDITOR
  www.addHandler(new SPIFFSEditor("admin", "nimda"));
#endif
#ifdef USE_PROFILING
 www.on("/profile", handleProfiling)
#endif
  www.serveStatic("/", SPIFFS, "/");
  www.onNotFound(handleNotFound);
#ifdef CORS_DEBUG
  DefaultHeaders::Instance().addHeader("Access-Control-Allow-Origin", "*");
  DefaultHeaders::Instance().addHeader("Access-Control-Allow-Headers", "content-type");
#endif
  www.begin();
}
